import abc

import openai
from pydantic import BaseModel

from .prompts import Template


class FunctionCall(BaseModel):
    name: str
    arguments: str


class Function(BaseModel):
    name: str
    description: str | None = None
    parameters: dict | None = None


class Message(BaseModel):
    role: str
    content: str
    name: str | None = None
    function_call: FunctionCall | None = None

    @classmethod
    def _from_template(
        cls,
        template: Template,
        role: str,
        name: str | None = None,
        function_call: FunctionCall | None = None,
        **kwargs,
    ):
        content = template.content.format(**kwargs)
        return cls(
            role=role,
            name=name,
            function_call=function_call,
            content=content,
        )


class UserMessage(Message):
    role: str = "user"

    @classmethod
    def from_template(
        cls,
        template: Template,
        name: str | None = None,
        function_call: FunctionCall | None = None,
        **kwargs,
    ):
        return cls._from_template(template, role="user", **kwargs)


class SystemMessage(Message):
    role: str = "system"

    @classmethod
    def from_template(
        cls,
        template: Template,
        name: str | None = None,
        function_call: FunctionCall | None = None,
        **kwargs,
    ):
        return cls._from_template(template, role="system", **kwargs)


class Choice(BaseModel):
    index: int
    message: Message
    finish_reason: str


class Usage(BaseModel):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class Completion(BaseModel):
    id: str
    object: str
    created: int
    model: str
    choices: list[Choice]
    usage: Usage | None


class LLM(BaseModel, abc.ABC):
    model: str
    api_key: str | None

    @abc.abstractmethod
    def create(self, **kwargs) -> Completion:
        pass  # pragma: no cover

    @abc.abstractmethod
    async def acreate(self, **kwargs):
        pass  # pragma: no cover


class OpenAIChat(LLM):
    messages: list[Message] = []
    function: Function | None = None
    function_call: FunctionCall | None = None
    temperature: float | None = None
    top_p: float | None = None
    n: int | None = None
    stream: bool | None = None
    stop: str | list[str] | None = None
    max_tokens: int | None = None
    presence_penalty: float | None = None
    frequency_penalty: float | None = None
    logit_bias: dict | None = None
    user: str | None = None

    def _serialize(self, **kwargs) -> dict:
        params = self.model_dump(exclude_none=True)
        params.update(kwargs)
        return params

    def create(self, **kwargs) -> Completion:
        params = self._serialize(**kwargs)
        completion = openai.ChatCompletion.create(**params)
        return Completion(
            id=completion.id,  # type: ignore
            object=completion.object,  # type: ignore
            created=completion.created,  # type: ignore
            model=completion.model,  # type: ignore
            choices=completion.choices,  # type: ignore
            usage=completion.usage,  # type: ignore
        )

    async def acreate(self, **kwargs):
        params = self._serialize(**kwargs)
        completion = await openai.ChatCompletion.acreate(**params)
        return Completion(
            id=completion.id,  # type: ignore
            object=completion.object,  # type: ignore
            created=completion.created,  # type: ignore
            model=completion.model,  # type: ignore
            choices=completion.choices,  # type: ignore
            usage=completion.usage,  # type: ignore
        )


OpenAIModels = [
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-0301",
    "gpt-3.5-turbo-0613",
    "gpt-4",
    "gpt-4-0314",
    "gpt-4-0613",
    "gpt-4-32k",
    "gpt-4-32k-0314",
    "gpt-4--32k-0613",
]

Models = OpenAIModels


def get(model: str, api_key: str | None = None) -> LLM:
    if model in OpenAIModels:
        return OpenAIChat(model=model, api_key=api_key)

    raise ValueError(f"Model {model} not found.")


def create_messages(
    *, system: str | None = None, user: str | None = None
) -> list[dict[str, str]]:
    messages = []
    if system is not None:
        messages.append(SystemMessage(content=system).model_dump(exclude_none=True))

    if user is not None:
        messages.append(UserMessage(content=user).model_dump(exclude_none=True))

    return messages



import string
from collections import UserDict
from pathlib import Path

import tomli
from pydantic import BaseModel


class Template(BaseModel):
    content: str

    def inputs(self) -> list[str]:
        """Return a list of field names in the contents."""
        formatter = string.Formatter()
        field_names = []
        for _, field_name, _, _ in formatter.parse(self.content):
            if field_name is not None:
                field_names.append(field_name)
        return sorted(list(set(field_names)))

    @classmethod
    def parse_toml(cls, filename: Path | str) -> "Template":
        """Parse a TOML file and return a Template object."""
        with open(filename, "rb") as f:
            toml_dict = tomli.load(f)
            return cls.model_validate(toml_dict)


class TemplateLibrary(UserDict):
    def add(self, name: str, template: Template) -> None:
        """Add a template to the library."""
        if not name:
            raise ValueError("Template must have a name.")
        self.data[name] = template

    @classmethod
    def from_directory(cls, dir: str | Path) -> "TemplateLibrary":
        """Load templates from a file or directory."""
        dir = Path(dir)
        if not dir.exists():
            raise FileNotFoundError(f"{dir} does not exist.")
        if not dir.is_dir():
            raise ValueError(f"{dir} is not a directory.")

        library = cls()
        for filename in dir.glob("**/*.json"):
            name = filename.parent / filename.stem
            bytes = filename.read_bytes()
            library.add(str(name), Template.model_validate_json(bytes))
        for filename in dir.glob("**/*.toml"):
            name = filename.parent / filename.stem
            library.add(str(name), Template.parse_toml(filename))
        return library
